const R = require('ramda');
const async = require('async');
const path = require('path');
const fs = require('fs');

const urlUtils = require('@src/utils/url-utils');
const createLambda = require('../utils/lambda-operation/create-lambda');
const updateLambda = require('../utils/lambda-operation/update-lambda');
const profileHelper = require('../utils/profile-helper');
const jsonUtility = require('../utils/json-utility');
const lambdaEndpointUtil = require('../utils/lambda-operation/lambda-endpoint-utility');
const CONSTANTS = require('../utils/constants');
const retrieveLambda = require('../utils/lambda-operation/retrieve-lambda');

module.exports = {
    deployLambda
};

function deployLambda(skillId, skillInfo, forceDeployment, profile, callback) {
    if (!skillId) {
        console.error('No skill ID found. Please first create skill, then deploy lambda.');
        process.exit(1);
        return;
    }
    const awsProfile = profileHelper.getAWSProfile(profile);
    if (!awsProfile) {
        console.warn(`[Warn]: Lambda deployment skipped because AWS credentials for profile [${profile}] are missing. CLI lambda functionalities 
can be enabled by running "ask init" again to add 'aws_profile' to ASK cli_config.`);
        if (typeof callback === 'function' && callback) {
            callback();
        }
        return;
    }

    _preMergeLambda(skillInfo.endpointsInfo, profile, (err, existingLambdaData) => {
        if (err) {
            callback(`[Error]: ${err}`);
            return;
        }
        // 1.merge endpoint information into the lambda resource config
        const lambdaResources = _mergeEndpointInfoToResource(skillInfo.endpointsInfo, existingLambdaData, profile);
        if (lambdaResources.length === 0) {
            console.log('[Info]: No lambda functions need to be deployed.');
            callback();
            return;
        }

        // 2.check if the Lambda runtime is nodejs8.10 if it uses ask-sdk
        const runtimeLessThenV8List = _checkRuntimeForAskSdkUsage(lambdaResources);
        if (runtimeLessThenV8List && runtimeLessThenV8List.length > 0) {
            console.warn('[Warn]: Please make sure the runtime of following Lambda resource(s) in project .ask/config '
                + 'is "nodejs8.10" since you are using ask-sdk as dependency.');
            runtimeLessThenV8List.forEach((lambdaMetadata) => {
                console.warn(`  ${lambdaMetadata.arn}`);
            });
            process.exit(1);
            return;
        }

        // 3.deploy lambda according to the state in lambda resources list
        async.eachSeries(lambdaResources, (lambdaResource, deployAsyncCallback) => {
            if (lambdaResource.state === CONSTANTS.LAMBDA.DEPLOY_STATE.CREATE) {
                createLambda.createLambda(skillId, skillInfo.skillName, lambdaResource, awsProfile, (createError, createResult) => {
                    if (createError) {
                        deployAsyncCallback(createError);
                    } else {
                        lambdaResource.arn = createResult.arn;
                        lambdaResource.revisionId = createResult.revisionId;
                        deployAsyncCallback();
                    }
                });
            } else if (lambdaResource.state === CONSTANTS.LAMBDA.DEPLOY_STATE.UPDATE) {
                updateLambda.updateLambda(lambdaResource, awsProfile, forceDeployment, (updateError, updateResult) => {
                    if (updateError) {
                        deployAsyncCallback(updateError);
                    } else {
                        lambdaResource.revisionId = updateResult.revisionId;
                        deployAsyncCallback();
                    }
                });
            } else {
                deployAsyncCallback(`[Error]: Failed to resolve the Lambda deploy state for ${lambdaResource.arn}. \
Please make sure this Lambda resource is correctly mapped to the existing alexaUsage in the config file.`);
            }
        }, (deployError) => {
            if (deployError) {
                console.error(deployError);
                process.exit(1);
                return;
            }
            // 4.save lambda resources to the config file and display results
            const stateDividedLambda = R.groupBy(lambdaResource => lambdaResource.state);
            const createLambdaList = stateDividedLambda(lambdaResources)[CONSTANTS.LAMBDA.DEPLOY_STATE.CREATE];
            const updateLambdaList = stateDividedLambda(lambdaResources)[CONSTANTS.LAMBDA.DEPLOY_STATE.UPDATE];
            _updateResourcesConfig(lambdaResources, profile);
            console.log('Lambda deployment finished.');
            _printLambdaDeployResults(createLambdaList, updateLambdaList);
            if (typeof callback === 'function' && callback) {
                callback(createLambdaList);
            }
            // 5.handle skill update
        });
    });
}

function _printLambdaDeployResults(createLambdaList, updateLambdaList) {
    if (createLambdaList && createLambdaList.length > 0) {
        console.log('Lambda function(s) created:');
        createLambdaList.forEach((metaData) => {
            console.log(`  [Lambda ARN] ${metaData.arn}`);
        });
    }
    if (updateLambdaList && updateLambdaList.length > 0) {
        console.log('Lambda function(s) updated:');
        updateLambdaList.forEach((metaData) => {
            console.log(`  [Lambda ARN] ${metaData.arn}`);
        });
    }
}

function _mergeEndpointInfoToResource(endpointsInfo, existingLambdaData, profile) {
    const lambdaResourcesList = jsonUtility.getPropertyValueFromObject(
        jsonUtility.read(path.join(process.cwd(), '.ask', 'config')),
        ['deploy_settings', profile, 'resources', 'lambda']
    ) || [];

    for (const domain of Object.keys(endpointsInfo)) {
        for (const region of Object.keys(endpointsInfo[domain])) {
            const usage = `${domain}/${region}`;
            const currentEndpoint = endpointsInfo[domain][region];

            if (!currentEndpoint.uri || !currentEndpoint.sourceDir) {
                // skip this endpoint usage when sourceDir or uri is not provided
                continue;
            }
            if (!fs.existsSync(currentEndpoint.sourceDir)) {
                console.error(`[Error]: The sourceDir (${currentEndpoint.sourceDir}) of the Lambda function ${currentEndpoint.uri} doesn't exist.`);
                process.exit(1);
                return null;
            }

            let isUsageExist = false;
            // 1.check if current usage exists in the alexaUsage in lambda resources list
            for (const lambdaResource of lambdaResourcesList) {
                // 1.1 usage is included in this lambda resource alexaUsage
                if (lambdaResource.alexaUsage.includes(usage)) {
                    if (urlUtils.isLambdaArn(lambdaResource.arn)) {
                        // 1.1.1 mark as update if this lambda resource already has lambda ARN
                        lambdaResource.state = CONSTANTS.LAMBDA.DEPLOY_STATE.UPDATE;
                        if (!lambdaResource.codeUri) {
                            lambdaResource.codeUri = currentEndpoint.sourceDir;
                        }
                        if (!lambdaResource.runtime && !lambdaResource.handler) {
                            // for backward compatibility use the following settings
                            lambdaResource.runtime = 'nodejs8.10';
                            lambdaResource.handler = 'index.handler';
                        }
                    } else if (!lambdaResource.arn && !lambdaResource.awsRegion) {
                        // 1.1.2 since awsRegion is generated internally, thus if awsRegion is null, then this resource hasn't been visited.
                        if (urlUtils.isLambdaArn(currentEndpoint.uri)) {
                            // 1.1.2.1 endpoint uri is an existing lambda ARN, mark as update
                            lambdaResource.state = CONSTANTS.LAMBDA.DEPLOY_STATE.UPDATE;
                            lambdaResource.arn = currentEndpoint.uri;
                            lambdaResource.awsRegion = lambdaEndpointUtil.parseLambdaArn(currentEndpoint.uri).awsRegion;
                            lambdaResource.codeUri = currentEndpoint.sourceDir;
                            if (!lambdaResource.functionName) {
                                lambdaResource.functionName = lambdaEndpointUtil.parseLambdaArn(currentEndpoint.uri).functionName;
                            }
                            if (!lambdaResource.runtime && !lambdaResource.handler) {
                                // for backward compatibility use the following settings
                                lambdaResource.runtime = 'nodejs8.10';
                                lambdaResource.handler = 'index.handler';
                            }
                        } else {
                            // 1.1.2.2 endpoint uri is function name to be created, mark as create
                            lambdaResource.state = CONSTANTS.LAMBDA.DEPLOY_STATE.CREATE;
                            lambdaResource.awsRegion = lambdaEndpointUtil.resolveLambdaAwsRegion(region);
                            lambdaResource.codeUri = currentEndpoint.sourceDir;
                            if (!lambdaResource.functionName) {
                                lambdaResource.functionName = currentEndpoint.uri;
                            }
                            if (!lambdaResource.runtime && !lambdaResource.handler) {
                                // for backward compatibility use the following settings
                                lambdaResource.runtime = 'nodejs8.10';
                                lambdaResource.handler = 'index.handler';
                            }
                        }
                    } // 1.1.3 else with arn and awsRegion, this Lambda resource has already be visited and no extra action
                    isUsageExist = true;
                    break;
                }
            }
            // 2.check when alexaUsage doesn't include current usage
            if (!isUsageExist) {
                // 2.1 when there is any lambda resource with same awsRegion and functionName, add the usage directly
                for (const lambdaResource of lambdaResourcesList) {
                    const resolvedRegion = urlUtils.isLambdaArn(currentEndpoint.uri)
                        ? lambdaEndpointUtil.parseLambdaArn(currentEndpoint.uri).awsRegion
                        : lambdaEndpointUtil.resolveLambdaAwsRegion(region);
                    if (lambdaResource.awsRegion === resolvedRegion && lambdaResource.arn === currentEndpoint.uri) {
                        lambdaResource.alexaUsage.push(usage);
                        isUsageExist = true;
                        break;
                    }
                }
                // 2.2 when the usage is really not used in resource
                if (!isUsageExist) {
                    if (urlUtils.isLambdaArn(currentEndpoint.uri)) {
                        // 2.2.1 update lambda when lambda ARN already exists
                        const existingLambdaResource = existingLambdaData[domain][region];
                        lambdaResourcesList.push({
                            state: CONSTANTS.LAMBDA.DEPLOY_STATE.UPDATE,
                            arn: currentEndpoint.uri,
                            functionName: lambdaEndpointUtil.parseLambdaArn(currentEndpoint.uri).functionName,
                            codeUri: currentEndpoint.sourceDir,
                            awsRegion: lambdaEndpointUtil.resolveLambdaAwsRegion(region),
                            runtime: existingLambdaResource.Configuration.Runtime,
                            handler: existingLambdaResource.Configuration.Handler,
                            alexaUsage: [usage],
                        });
                    } else {
                        // 2.2.2 create new lambda when the lambda resource doesn't have its resource config (for backward compatibility)
                        lambdaResourcesList.push({
                            state: CONSTANTS.LAMBDA.DEPLOY_STATE.CREATE,
                            codeUri: currentEndpoint.sourceDir,
                            awsRegion: lambdaEndpointUtil.resolveLambdaAwsRegion(region),
                            functionName: currentEndpoint.uri,
                            runtime: 'nodejs8.10',
                            handler: 'index.handler',
                            alexaUsage: [usage]
                        });
                    }
                }
            }
        }
    }
    return lambdaResourcesList;
}

function _checkRuntimeForAskSdkUsage(lambdaResources) {
    const runtimeLessThenV8List = [];
    for (const lambda of lambdaResources) {
        if (lambdaEndpointUtil.isLambdaUsingAskSdk(lambda.codeUri)
            && (lambda.runtime.startsWith('nodejs') && lambda.runtime !== 'nodejs8.10')) {
            runtimeLessThenV8List.push(lambda);
        }
    }
    return runtimeLessThenV8List;
}

function _updateResourcesConfig(lambdaResources, profile) {
    // filter unnecessary properties
    for (const lambda of lambdaResources) {
        lambda.state = undefined;
        if (lambda.customName) {
            lambda.customName = undefined;
        }
    }
    const keySortedLambdaResources = [];
    for (const lambda of lambdaResources) {
        const tempLambda = {};
        R.keys(lambda).sort().forEach((key) => {
            tempLambda[key] = lambda[key];
        });
        keySortedLambdaResources.push(tempLambda);
    }
    jsonUtility.writeToProperty(
        path.join(process.cwd(), '.ask', 'config'),
        ['deploy_settings', profile, 'resources', 'lambda'],
        keySortedLambdaResources
    );
}

function _preMergeLambda(endpointsInfo, profile, callback) {
    const result = {};
    const awsProfile = profileHelper.getAWSProfile(profile);

    async.forEach(Object.keys(endpointsInfo), (domain, domainLevelCallback) => {
        result[domain] = {};
        async.forEach(Object.keys(endpointsInfo[domain]), (region, regionLevelCallback) => {
            const { uri } = endpointsInfo[domain][region];
            if (!urlUtils.isLambdaArn(uri)) {
                regionLevelCallback();
                return;
            }
            const { awsRegion } = lambdaEndpointUtil.parseLambdaArn(uri);
            retrieveLambda.retrieveLambdaFunction(uri, awsRegion, awsProfile, (err, functionData) => {
                if (err) {
                    regionLevelCallback(err);
                    return;
                }
                // exclude no lambda found case
                if (functionData) {
                    result[domain][region] = functionData;
                }
                regionLevelCallback();
            });
        }, (regionError) => {
            domainLevelCallback(regionError);
        });
    }, (domainError) => {
        callback(domainError, domainError ? null : result);
    });
}
