'use strict';

const expect = require('chai').expect;
const sinon = require('sinon');
const fs = require('fs');
const path = require('path');
const shell = require('shelljs');
const mkdirp = require('mkdirp');
const jsonfile = require('jsonfile');
const HookManager = require('@src/utils/hook-manager');
const jsonUtility = require('@src/utils/json-utility');
const httpClient = require('@src/clients/http-client');
const displayUtil = require('@src/utils/display-utility');

const noop = () => {};

describe('Hook Manager test', () => {
    const TEST_SKILL_NAME = 'testSkillName';
    const TEST_TARGET = 'test-all';
    const TEST_DEBUG = false
    const SCRIPT_LIST = ['pre_deploy_hook.sh'];
    const TEST_ERROR = 'testError';
    const TEST_RUN_TIME = 'node';
    const TEST_PROFILE = 'testProfile';


    describe('# inspect correctness for constructor', () => {
        beforeEach(() => {
            sinon.stub(HookManager.HookManager.prototype, 'validate');
            sinon.stub(process, 'cwd').callsFake(() => {
                return './';
            });
        });
            
        it('| Constructor setup for post new hook', () => {
            //setup
            const TEST_CATEGORY = 'post_new';    
            const TEST_OS_TYPE = 'WINDOWS'; 
            const TEST_RUN_TIME = 'node';
            let sandbox = sinon.createSandbox();
            sandbox.stub(process,'platform').value(TEST_OS_TYPE);

            const test = {
                category: TEST_CATEGORY,
                skillName: TEST_SKILL_NAME,
                doDebug:TEST_DEBUG,
                ostype: TEST_OS_TYPE,
                target: TEST_TARGET,
                hookPath: `${TEST_SKILL_NAME}/hooks`,
                scriptPath: `${TEST_SKILL_NAME}/hooks/${SCRIPT_LIST[0]}`
            }       
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {
                //verify
                expect(err).to.be.null;
                expect(hook).deep.equal(test);
            });
            //clear
            sandbox.restore();     
        });

        it('| Constructor setup for pre deploy hook', () => {
            //setup
            const TEST_CATEGORY = 'pre_deploy';    
            const TEST_OS_TYPE = 'LINUX'; 
            let sandbox = sinon.createSandbox();
            sandbox.stub(process,'platform').value(TEST_OS_TYPE);
            const test = {
                category: TEST_CATEGORY,
                skillName: TEST_SKILL_NAME,
                doDebug:TEST_DEBUG,
                ostype: TEST_OS_TYPE,
                target: TEST_TARGET,
                hookPath: 'hooks',
                scriptPath: `hooks/${SCRIPT_LIST[1]}`
            }       
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {
                //verify
                expect(err).to.be.null;
                expect(hook).deep.equal(test);
            });
            //clear
           sandbox.restore();
        })

        afterEach(() => {
            HookManager.HookManager.prototype.validate.restore();
            process.cwd.restore();
        })

    });

    describe('# getSourceDir function check', () => {
        let sandbox;
        const TEST_CATEGORY = 'post_new';
        beforeEach(() => {
            sandbox = sinon.createSandbox();
            sandbox.stub(jsonfile, 'readFileSync');
            sandbox.stub(HookManager.HookManager.prototype,'validate');
            sandbox.stub(HookManager.HookManager.prototype,'_getRunTime');
        });

        it('| getSourceDir parses skill.json file and returns sourceDir location', () => {
            //setup
            sandbox.stub(jsonUtility, 'getPropertyValueFromObject').callsFake(() => {
                return {
                    "smartHome": {
                        "endpoint": {
                        },
                    },
                    "custom": {
                        "endpoint": {
                          "sourceDir": "test"
                        }
                      }
                    };
            });
            HookManager.HookManager.prototype.validate.callsArgOnWith(0, {}, null, true);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, null, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (_err, hook) => {       
                const sourceDir = hook._getSourceDir();
                //verify
                expect(sourceDir).equal('test');
            });
        });

        it('| getSourceDir returns null if sourceDir is missing in skill.json file', () => {
            //setup
            sandbox.stub(jsonUtility, 'getPropertyValueFromObject').callsFake(() => {});
            HookManager.HookManager.prototype.validate.callsArgOnWith(0, {}, null, true);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, null, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {       
                const sourceDir  = hook._getSourceDir();
                //verify
                expect(sourceDir).to.be.null;
            });
        });

        afterEach(() => {
            sandbox.restore();
        });
    });

    describe('# getRunTime function check' ,() => {
        let sandbox;
        beforeEach(() => {
            sandbox = sinon.createSandbox();
            sandbox.stub(path, 'join');
            sandbox.stub(jsonfile, 'readFileSync');
            sandbox.stub(HookManager.HookManager.prototype,'validate');
        });

        it('| getRuntime identifies multiple runtime and returns null', () => {
            //setup
            const TEST_CATEGORY = 'post_new';   
            sandbox.stub(console, 'warn');
            sandbox.stub(jsonUtility, 'getPropertyValueFromObject').callsFake(() => {
                return [
                    {
                      "runtime": "nodejs8.10",
                    },
                    {
                        "runtime": "python2.7",
                    }
                  ];
            });
            HookManager.HookManager.prototype.validate.callsArgOnWith(0, {}, null, true);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, null, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {       
                expect(console.warn.args[0][0]).equal('[Warn]: Different lambda runtime found. CLI currently does not support this functionality.'
                + 'Developers should either work on their own hook scripts to build lambda for different language. '
                + 'Or use only one languages for the skill.\n');
                expect(hook.runTime).to.be.null;
            });
            //clear
        });

        it('| getRuntime identifies multiple runtime and validates all to be similar, returns success', () => {
            //setup
            const TEST_CATEGORY = 'post_new';   
            sandbox.stub(jsonUtility, 'getPropertyValueFromObject').callsFake(() => {
                return [
                    {
                      "runtime": "nodejs8.10",
                    },
                    {
                        "runtime": "nodejs6.10",
                    }
                  ];
            });
            HookManager.HookManager.prototype.validate.callsArgOnWith(0, {}, null, true);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, null, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {       
                expect(err).to.be.null;
                expect(hook.runTime).equal('nodejs');
            });
            //clear
        });

        it('| getRuntime checks for backward compatible node templates, returns node as runTime', () => {
            //setup
            const TEST_CATEGORY = 'post_new';   
            sandbox.stub(jsonUtility, 'getPropertyValueFromObject').callsFake(() => null);
            sandbox.stub(HookManager.HookManager.prototype,'_getSourceDir').callsFake(() => 'test');
            sandbox.stub(fs, 'existsSync').returns(true);

            HookManager.HookManager.prototype.validate.callsArgOnWith(0, {}, null, true);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, null, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {       
                expect(err).to.be.null;
                expect(hook.runTime).equal('node');
            });
            //clear
        });

        it('| getRuntime returns null for missing sourceDir info in skill.json file for backward compatability', () => {
            //setup
            const TEST_CATEGORY = 'post_new';   
            sandbox.stub(jsonUtility, 'getPropertyValueFromObject').callsFake(() => null);
            sandbox.stub(HookManager.HookManager.prototype,'_getSourceDir').callsFake(() => null);
            sandbox.stub(fs, 'existsSync').returns(true);

            HookManager.HookManager.prototype.validate.callsArgOnWith(0, {}, null, true);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, null, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {       
                expect(err).to.be.null;
                expect(hook.runTime).to.be.null;
            });
            //clear
        });

        it('| getRuntime checks for backward compatible node templates, returns python as runTime', () => {
            //setup
            const TEST_CATEGORY = 'post_new';   
            sandbox.stub(console, 'warn');
            sandbox.stub(jsonUtility, 'getPropertyValueFromObject').callsFake(() => null);
            sandbox.stub(HookManager.HookManager.prototype,'_getSourceDir').callsFake(() => 'test');
            sandbox.stub(fs, 'existsSync').returns(false);

            HookManager.HookManager.prototype.validate.callsArgOnWith(0, {}, null, true);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, null, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {       
                expect(err).to.be.null;
                expect(hook.runTime).equal('python');
            });
            //clear
        });
        
        afterEach(() => {
            sandbox.restore();
        });

    });

    describe('# Validate function check', () => {
        beforeEach(() => {
            sinon.stub(process, 'cwd').callsFake(() => {
                return `${TEST_SKILL_NAME}`;
            });
        });

        it('| Check if hook folder exists, returns callback true value', () => {
            //setup
            sinon.stub(fs, 'existsSync').returns(true);
            const TEST_CATEGORY = 'post_new';    
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {       
                expect(err).to.be.null;
                expect(hook).to.be.not.null;
            });
            //clear
            fs.existsSync.restore();
        });

        it('| Skill directory is inconsistent with skillname, expects callback error', () => {
            //setup
            const TEST_CATEGORY = 'pre_deploy';
            const TEST_SKILL_NAME = 'test';
            sinon.stub(fs, 'existsSync').returns(false);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {
                //verify
                expect(err).equal(`[Error]: Directory name is not consistent with the skill name present in skill.json: ${TEST_SKILL_NAME}. Please change your project directory name accordingly.`)
                expect(hook).to.be.null;
            });
            //clear
            fs.existsSync.restore();

        });

        it('| Validate returns sucessfully for dir consistency and hook exits, expects callback success', () => {
            //setup
            const TEST_CATEGORY = 'pre_deploy';
            sinon.stub(fs, 'existsSync').returns(true);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {
                //verify
                expect(err).to.be.null;
                expect(hook).to.be.not.null;
            });
            //clear
            fs.existsSync.restore();
        });
        afterEach(() => {
            process.cwd.restore();
        });
    });

    describe('# createHook check', ()=> {
        beforeEach(() => {
            sinon.stub(process, 'cwd').callsFake(() => {
                return `${TEST_SKILL_NAME}`;
            });
            sinon.stub(fs, 'existsSync').returns(false);
            sinon.stub(mkdirp, 'sync');
            sinon.stub(fs, 'writeFileSync');
            sinon.stub(console, 'info');
        });

        it('| Creating hook folder returns error, expect console error', () => {
            //setup
            sinon.stub(console,'error');

            mkdirp.sync.callsFake(() => { throw 'Dir Error';});
            const TEST_CATEGORY = 'pre_deploy'; 
            const TEST_PATH = path.join( TEST_SKILL_NAME, 'hooks'); 
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {
                //verify
                expect(console.error.args[0][0]).equal(`[Error]: Cannot create directory path for: ${TEST_PATH}`);
                expect(err).to.be.not.null;
                expect(hook).to.be.null;
            });
            //clear
            console.error.restore();
        
        });

        it('| _cloneHookScripts returns error, expect callback error', () => {
            //setup
            const TEST_CATEGORY = 'post_new';
            sinon.stub(HookManager.HookManager.prototype, '_cloneHookScripts');;
            HookManager.HookManager.prototype._cloneHookScripts.callsArgOnWith(2, {}, TEST_ERROR);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {
                //verify
                expect(err).equal(TEST_ERROR)
                expect(hook).to.be.null;
            });
            //clear
            HookManager.HookManager.prototype._cloneHookScripts.restore();
        });

        it('| helper _cloneHookScripts downloads hooks successfully for older nodejs skills, expects success', () => {
            //setup
            const TEST_CATEGORY = 'post_new';  
            sinon.stub(HookManager.HookManager.prototype, '_cloneHookScripts');;
            HookManager.HookManager.prototype._cloneHookScripts.callsArgOnWith(2, {}, null);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {
                //verify
                expect(err).to.be.null
                expect(hook).to.be.not.null;
            });
            //clear
            HookManager.HookManager.prototype._cloneHookScripts.restore();
        });

        it('| empty hook script created for python cloned skills, expect success', () => {
            const TEST_CATEGORY = 'post_new';  
            const TEST_RUN_TIME = 'python2.7'
            sinon.stub(path, 'join').returns(TEST_SKILL_NAME);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (err, hook) => {
                //verify
                expect(fs.writeFileSync.args[0][0]).equal(TEST_SKILL_NAME);
                expect(fs.writeFileSync.calledOnce).to.be.true;
            });
            //clear
            path.join.restore();
        });

        afterEach(() => {
            mkdirp.sync.restore();
            fs.writeFileSync.restore();
            process.cwd.restore();
            fs.existsSync.restore();
            console.info.restore();
        })
    });

    describe('# execute check', () => {
        beforeEach(() => {
            sinon.stub(HookManager.HookManager.prototype, 'validate');    
        });

        it('| shell execution fails, expect callback error', () => {
            //setup
            sinon.stub(shell, 'exec').callsFake(() => {
                return {
                    code : 1
                }
            });
            //call
            const TEST_CATEGORY = 'pre_deploy';
            HookManager.HookManager.prototype.validate.callsArgOnWith(0, {}, null, true);
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (_err, hook) => {
                hook.execute((error, res) => {
                    //verify
                    expect(error).equal('[Error]: Hook Scripts failed');
                    expect(res).to.be.null;
                });
            });
            //clear
            shell.exec.restore();
        });

        it('| hook script executes successfully, expect sucess callback', () => {
            //setup
            sinon.stub(shell, 'exec').callsFake(() => {
                return {
                    code : 0
                }
            });
            //call
            const TEST_CATEGORY = 'pre_deploy';
            HookManager.HookManager.prototype.validate.callsArgOnWith(0, {}, null, true);
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (_err, hook) => {
                hook.execute((error, res) => {
                    //verify
                    expect(error).to.be.null;
                    expect(res).equal(`Successful executed ${hook.category} hook script`);
                });
            });
            //clear
            shell.exec.restore();
        });

        afterEach(() => {
            HookManager.HookManager.prototype.validate.restore();
            
        });
    });

    describe('# getHookScripts check',() => {
        let sandbox;
        const TEST_CATEGORY = 'post_new';
        beforeEach(() => {
            sandbox = sinon.createSandbox();
            sandbox.stub(HookManager.HookManager.prototype, 'validate');
            sandbox.stub(httpClient, 'request');
            HookManager.HookManager.prototype.validate.callsArgOnWith(0, {}, null, true);
        });

        it('| httpClient request gets called successfully', () => {
            //setup
            const TEST_URL = 'https://test.com'
            httpClient.request.callsArgOnWith(3, {}, null, { body : 'test'});
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (_err, hook) => {
                hook._getHookScripts(TEST_URL, (error, res) => {
                    //verify
                    expect(error).to.be.null;
                    expect(res).equal('test');
                });
            });
            //clear
        });
        
        it('| httpClient request returns error, expect error callback', () => {
            //setup
            const TEST_URL = 'https://test.com'
            httpClient.request.callsArgOnWith(3, {}, TEST_ERROR, null);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (_err, hook) => {
                hook._getHookScripts(TEST_URL, (error, res) => {
                    //verify
                    expect(error).equal(TEST_ERROR);
                    expect(res).to.be.null;
                });
            });
            //clear
        });

        it('| httpClient request returns null status code, expect error callback ', () => {
            //setup
            const TEST_URL = 'https://test.com'
            httpClient.request.callsArgOnWith(3, {}, null, { statusCode : null});
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (_err, hook) => {
                hook._getHookScripts(TEST_URL, (_error, res) => {
                    //verify
                    expect(res).to.be.null;
                });
            });
            //clear
        });

        afterEach(() => {
            sandbox.restore();
        });
    });

    describe('# _cloneHookScripts check', () => {
        const TEST_CATEGORY = 'post_new';  
        let sandbox;
        beforeEach(() => {
            const spinnerSpy = sinon.stub(displayUtil,"Spinner");    
            spinnerSpy.prototype.start = sinon.stub().returns(null);
            spinnerSpy.prototype.stop = sinon.stub().returns(null);
            sandbox = sinon.createSandbox();
            sandbox.stub(mkdirp, 'sync');
            sandbox.stub(fs, 'writeFileSync');
            sandbox.stub(HookManager.HookManager.prototype, 'validate');
            sandbox.stub(HookManager.HookManager.prototype, '_getHookScripts');
            HookManager.HookManager.prototype.validate.callsArgOnWith(0, {}, null, true);
        });

        it('| getHookScripts successfully downloads hooks and returns success',() => {  
            //setup     
            const TEST_SCRIPT_PATH = 'https://test.com';
            HookManager.HookManager.prototype._getHookScripts.callsArgOnWith(1, {}, null, TEST_SKILL_NAME);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (_err, hook) => {
                hook._cloneHookScripts(TEST_SCRIPT_PATH, SCRIPT_LIST, (err) => {
                    //verify
                    expect(err).to.be.null;
                    expect(fs.writeFileSync.calledOnce).to.be.true;
                    expect(fs.writeFileSync.args[0][1]).equal(TEST_SKILL_NAME);
                });
            });
            //clear
        });

        it('| getHookScripts returns error, expect callback error',() => {  
            //setup     
            const TEST_SCRIPT_PATH = 'https://test.com';
            HookManager.HookManager.prototype._getHookScripts.callsArgOnWith(1, {}, TEST_ERROR, null);
            //call
            HookManager.createManager(TEST_CATEGORY, TEST_SKILL_NAME, TEST_RUN_TIME, TEST_PROFILE, TEST_DEBUG, TEST_TARGET, (_err, hook) => {
                hook._cloneHookScripts(TEST_SCRIPT_PATH, SCRIPT_LIST, (err) => {
                    //verify
                    expect(err).equal(TEST_ERROR);
                    expect(fs.writeFileSync.calledOnce).to.be.false;
                });
            });
            //clear
        });

        afterEach(() => {
            displayUtil.Spinner.restore();  
            sandbox.restore();
        });
    });
});
